{{
Fast single encoder object by Graham Stabler 11th June 2007

The read-process-update loop takes aproximately 760ns, 1.31Mhz. 

Code based on a method suggested by "Skogsgurra" and ellaborated on by Chip Gracey.

How it works:

The A and B phase from the encoder are combined with the previous two readings to make a 4-bit number.
This number provides an index to a table containing 0's,1's and -1's.  If the combination of old and new
readings means no rotation a zero is found in the table, if the rotation is clockwise a 1 is found and if
anticlockwise a -1 is found.

These can then be simply added to the position register used.

In this instance the lookup table is contained in a single 32-bit long, there are 16 pairs of numbers, one for
each possible combination (note encoders cannot produce all 16 combinations of a 4-bit number).

To produce an index the 4-bit number is multiplied by 2 (shift once left) and then used as the argument for a
left shift performed ona copy of the table.  Note that element 15 in the table is in bits 0:1 so they are shifted 30 times
into 30:31.  Then an arithmetic right shift is performed, this ensures that the 2-bit negative number retains
its sign because the msb is extended during the shift:

11000000_00000000_00000000_00000000

becomes

11111111_11111111_11111111_11111111 Or minus one!

Simple!


(Table based on Jeff Martin's quadrature code)

'-------------|-------------|-----------------|----------------|-----------
'             |             |                 |   2-bit table  |  Meaning
'old -> new   |   4-bit     |  Index in table |     content    |
'-------------|-------------|-----------------|----------------|-----------
' 00  ->  00  |    0000     |        0        |      00        |
' 01  ->  01  |    0101     |        5        |      00        |   No movement 
' 11  ->  11  |    1111     |        15       |      00        |
' 10  ->  10  |    1010     |        10       |      00        |
'-------------|-------------|-----------------|----------------|-----------
' 00  ->  01  |    0001     |        1        |      01        |
' 01  ->  11  |    0111     |        7        |      01        |
' 11  ->  10  |    1110     |        14       |      01        |   Clockwise
' 10  ->  00  |    1000     |        8        |      01        |
'-------------|-------------|-----------------|----------------|-----------
' 00  ->  10  |    0010     |        2        |      11   (-1) |
' 10  ->  11  |    1011     |        11       |      11   (-1) |   Anticlockwise
' 11  ->  01  |    1101     |        13       |      11   (-1) |
' 01  ->  00  |    0100     |        4        |      11   (-1) |

_
}}


VAR

  long  cog
  long  pos
  
PUB start(pin,PosAddr) : okay | pmask

  _pin           := pin
  pos            := PosAddr
  okay := cog := cognew(@entry, pos) + 1

PUB stop

  if cog
    cogstop(cog~ - 1) 

DAT

'************************************************************
'* Reading an encoder the Skogsgurra and Chip Gracey way    *
'************************************************************
                        
entry                   mov   combined, #0
                        mov   p, par
                        mov   pinmask,#3           ' Create pinmask
                        shl   pinmask,_pin                                                  
                                                
:loop                   mov   look, lookyup        ' load look up table
                        mov   state, ina           ' Read inputs
                        and   state, pinmask       ' Mask
                        shr   state, _pin           ' shift into bits 0 and 1              
                        shl   combined,#2          ' Shift old bits left
                        or    combined,state       ' combine to provide 4 bit word
                        and   combined,#15         ' mask off really old bits
                        mov   shift,combined       ' Make ready to use as index 
                        shl   shift,#1             ' Shift to multiply by two
                        shl   look,shift           ' Use to shift look up table bits to bits 30,31                                       
                        sar   look,#30             ' Shift to bits 0 and 1 keeping the sign                        
                        
                        adds  count,look           ' Add the result to the count

                        wrlong count,p           '          
                        jmp   #:loop 

count                   long    0
combined                long    0
pinmask                 long    0

'                                 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
lookyup                 long    %00_01_11_00_11_00_00_01_01_00_00_11_00_11_01_00
_pin                    long    0  

state                   res     1
look                    res     1
shift                   res     1
p                       res     1
